
\
\ delay.c:  Delay calibration, the Linux way
\
\ By Ramon van Handel (19/12/98)
\ Based on the method used in linux 2.0.30 init/main.c
\
\ Resembles a lot the way I thought up in my original email, just a tiny
\ bit more accurate.  The principle is the same, implementation is
\ different.  I didn't make this as detailed as linux does, but I
\ documented it pretty well.  If you really want to trim this to the
\ microsecond, have a look at the linux sources.  The main calibration
\ code (init/main.c) is similar, but have a look in include/linux/delay.h,
\ include/asm-i386/delay.h, and arch/i386/lib/delay.S.
\                                                        -- Ramon

\ *************************************************************************


\ Allright, let's review how we're going to do this.
\
\ Goal:
\ The goal is to create a reasonably accurate version of the delay()
\ function, which creates a delay in multiples of one millisecond.
\
\ Restrictions:
\ We don't want to use the PIT for the delay.  In your average
\ bare-hardware system (ie OS) you'll be using the PIT for quite a few
\ other things already (in my OS code, channel 0 and 1 are used by the
\ scheduler, and channel 2 would probably be needed to control the
\ speaker.)
\
\ Method:
\ A simple loop would delay the computer wonderfully.  Ie, take something
\ like:
\
\         for(i=0;i<BIGNUMBER;i++);  // Delay loop
\
\ That kills time wonderfully.  All we need to do is to find out what
\ BIGNUMBER we need to use to delay one millisecond.  Finding the correct
\ BIGNUMBER for your machine is called *delay loop calibration*.
\
\ We can calibrate the delay loop using the PIT.  In the initialisation
\ phase of the OS, when the PIT is not yet hooked to the scheduler, it
\ can be freely used to do this.  When the delay loop has been calibrated
\ against the PIT, the PIT is free to be used for other purposes.
\
\ There are many ways to calibrate a delay loop with the PIT.  This is
\ one way (used in Linux.)  More documentation in the code.

\ *************************************************************************

\ The delay function:
\ Have a look at this function before you look at the calibration code.
\ Basically, delay_count is the number you have to count to in order to
\ kill 1ms.  delay() takes the amount of milliseconds to delay as
\ parameter.
\
\ The goal of the rest of this code is to determine what delay_count *is*
\ for your machine.
\
\ Keep delay_count initialised to 1.  We'll need that later.


1 VALUE DELAY_COUNT


\ When I was testing the delay code, I notice that the calibration we use
\ is very delicate.  Originally, I had inlined the delay for() loops into
\ the delay() and calibrateDelayLoop() functions, but this doesn't work.
\ Due to different register allocation or something similar the compiler
\ might have generated different code in the two situations (I didn't
\ check this.)  Or alignment differences might have caused the problems.
\ Anyway, to solve the problem, we always use the __delay() function for
\ the delay for() loop, because it always is the same code with the same
\ alignment. __delay() is called from delay() as well as from
\ calibrateDelayLoop().  By using __delay() we can fine-tune our
\ calibration without it losing its finesse afterwards.


: _DELAY ( loops -- )
  0 DO LOOP
;

: _MS ( ms -- )
  DELAY_COUNT * _DELAY
;

\ *************************************************************************

\ The timer ISR:
\ In order to calibrate the delay loop, we program the PIT channel 0 with
\ a fixed timer rate of MILLISEC milliseconds.  The only thing the ISR
\ does is increment a timer, ticks, on every interrupt (that is, every
\ 10ms in this case.)  The main program will use this counter in order to
\ calibrate the delay loop.
 1000
 10 DUP CONSTANT MILLISEC   \ Take a 10ms fixed timer rate   
 /      CONSTANT FREQ       \ Timer freqency for MILLISEC    

0 VALUE TICKS

:  PIT_DELAY_ISR >INT TICKS 1+ TO TICKS  INT> ;

\ ************************************************************************


\ Calibrate the delay loop:
\ This is the most important part of the whole business.  I'll explain
\ the general strategy here, which is supplemented by comments in the
\ code.
\
\ Stage 1:
\ First, we are going to make a rough approximation of delay_count. We
\ are going to try to determine between which two powers of 2 delay_count
\ needs to be.  In order to do that, we first wait until the next timer
\ tick has started (we know it has when the ticks variable has just been
\ incremented.)  Then we start a delay with delay_count of 1.  When the
\ delay has finished, we look whether a timer tick has passed during the
\ delay (nah !).  If not, we shift delay_count to the right once and try
\ again.  We continue until after the delay, one timer tick has passed.
\ We then know that our value for delay_count is too big, so the 'real'
\ delay_count is between (delay_count) and (delay_count>>1).
\
\ Stage 2:
\ Afterwards, we start refining our value.  We shift our delay_count to
\ the right once to obtain out bottom-value for delay_count (ie we then
\ know that the 'real' delay_count is bigger than the actual
\ delay_count.)  delay_count is a power of 2, so only one bit is set.
\ What we do is, we start setting bits in delay_count and see whether
\ delay_count is bigger than a tick afterwards. An example helps
\ clarifying matters:
\
\ Imagine we found in stage 1 that delay_count is between 00100000b and
\ 01000000b.  (In reality it would be much bigger, but for the example
\ this will do.)  Before we start our fine-calibration, we set
\ delay_count to 00100000b.
\ Now we turn on the bit after the set bit, ie:  00110000b
\ We do the delay like we did previously, and see whether ticks has
\ changed during it.  If it has, we know that the value is too big, and
\ turn the bit back off.  If it hasn't, we leave it on.  In either case,
\ we proceed with the next bit:
\
\     00100000b   <--- start value
\     00110000b   <--- tick hasn't passed
\     00111000b   <--- tick has passed, turn it off: 00110000b
\     00110100b   <--- tick has passed, turn it off: 00110000b
\     00110010b   <--- tick hasn't passed
\     00110011b   <--- tick hasn't passed
\
\ And there is our delay_count value.
\
\ In order to save time (calibrating can be a time-consuming matter, and
\ tuning it extremely finely is often useless because at a certain level
\ you get additional delay by calling the delay() function, and returning
\ from it, etc.) we only calculate delay_count up to a certain precision
\ (ie, we only calculate PRECISION extra bits after stage 1.)  A
\ PRECISION value of 8 gives good results (according to the linux
\ sources, it gives an imprecision of approximately 1%.)

8 CONSTANT PRECISION                 \ Calibration precision

0x30 CONSTANT TMR_BOTH 
0x06 CONSTANT TMR_MD3
0x43 CONSTANT TMR_CTRL
: _INIT_PIT ( temp  channel -- )
           || temp  channel ||
           (( temp  channel ))
    channel 0x40 *  TMR_BOTH + TMR_MD3 +  TMR_CTRL PC!
    temp         channel 0x40 + PC!
   temp 8 RSHIFT channel 0x40 + PC!
;

M_VEC CONSTANT M_VEC
\ calibrateDelayLoop

: CDL
    1193180 FREQ / 0 _INIT_PIT
    0x8E00   [']  PIT_DELAY_ISR  M_VEC  SET_VECTOR
    0x21 PC@ 1 INVERT AND 0x21 PC!
    \ Stage 1:  Coarse calibration
    \ Initialise timer interrupt with MILLISECOND ms interval
    BEGIN
        DELAY_COUNT 2*          \ Next delay count to try
     TO DELAY_COUNT
        TICKS                   \ Wait for the start of the next
        BEGIN  DUP TICKS <>     \ timer tick
        UNTIL DROP

        TICKS                   \ Start measurement now
        DELAY_COUNT _DELAY       \ Do the delay
        TICKS <>
    UNTIL                       \ Until delay is just too big

    DELAY_COUNT 2/              \ Get bottom value for delay
     TO DELAY_COUNT

    \ Stage 2:  Fine calibration

    DELAY_COUNT                 \ Which bit are we going to test

    PRECISION 0
    DO  2/ DUP IF LEAVE THEN        \ Next bit to calibrate          
                                    \ If we have done all bits, stop 

        DUP DELAY_COUNT OR    \ Set the bit in delay_count     
         TO DELAY_COUNT

        TICKS                   \ Wait for the start of the next 
        BEGIN  DUP TICKS <>     \ timer tick                     
        UNTIL DROP
        TICKS                    \ Start measurement now          
        DELAY_COUNT _DELAY       \ Do the delay                   
        TICKS <>
        IF  DUP INVERT DELAY_COUNT         \ If a tick has passed, turn the 
               AND  TO DELAY_COUNT         \ calibrated bit back off 
        THEN
    LOOP DROP

    \ We're finished:  Do the finishing touches

    0x21 PC@ 1 OR 0x21 PC!        \ Our wonderful PIT can stop now  

    DELAY_COUNT  MILLISEC /       \ Calculate delay_count for 1ms   
  TO DELAY_COUNT
;
: PIT_HANDLER
  [ 0xCD C, 0x1C C, ] \ INT 1CH
  0x20 0x20 PC!
;

: PIT_ISR >INT PIT_HANDLER INT> ;

: INIT_PIT
 ( 1193180/18.2 ) 65559 0 _INIT_PIT
  0x8E00 ['] PIT_ISR  M_VEC  SET_VECTOR
;

